#include <string.h>
#include <stdlib.h>

#include "vdc.h"
#include "vdc_drawing.h"

// Clear the LCD.
void vdc_clear(void) {
	memset(vdc_frame_buffer, 0, sizeof(vdc_frame_buffer));
}

// Set a pixel on the LCD.
void vdc_set_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		vdc_frame_buffer[y * 20 + x / 16] |= (0x8000 >> (x & 15));
	}
}

// Clear a pixel on the LCD.
void vdc_clear_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		vdc_frame_buffer[y * 20 + x / 16] &= ~(0x8000 >> (x & 15));
	}
}

// Invert a pixel on the LCD.
void vdc_invert_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		vdc_frame_buffer[y * 20 + x / 16] ^= (0x8000 >> (x & 15));
	}
}

// Draw a pixel on the LCD using the current plotting mode.
void (*vdc_draw_pixel)(int, int) = &vdc_set_pixel;

// Set the current plotting mode.
void vdc_set_plot_mode(enum VDC_PLOT_MODE mode) {
	switch (mode) {
		case VDC_PLOT_MODE_SET:
			vdc_draw_pixel = &vdc_set_pixel;
			break;
		case VDC_PLOT_MODE_CLEAR:
			vdc_draw_pixel = &vdc_clear_pixel;
			break;
		case VDC_PLOT_MODE_INVERT:
			vdc_draw_pixel = &vdc_invert_pixel;
			break;

	}
}

// Draw a line on the LCD.
// Adapted from http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
void vdc_draw_line(int x0, int y0, int x1, int y1) {
	int deltax, deltay;
	int error, ystep;
	int x, y;
	unsigned char steep = abs(y1 - y0) > abs(x1 - x0);
	if (steep) {
		int s;
		s = x0; x0 = y0; y0 = s;
		s = x1; x1 = y1; y1 = s;
	}
	if (x0 > x1) {
		int s;
		s = x0; x0 = x1; x1 = s;
		s = y0; y0 = y1; y1 = s;
	}
	deltax = x1 - x0;
	deltay = abs(y1 - y0);
	error = deltax / 2;
	y = y0;
	if (y0 < y1) {
		ystep = 1;
	} else {
		ystep = -1;
	}
	for (x = x0; x <= x1; ++x) {
		if (steep) {
			vdc_draw_pixel(y, x);
		} else {
			vdc_draw_pixel(x, y);
		}
		error = error - deltay;
		if (error < 0) {
			y = y + ystep;
			error = error + deltax;
		}
	}
}

// Draw a circle on the LCD.
// Adapted from http://en.wikipedia.org/wiki/Midpoint_circle_algorithm
void vdc_draw_circle(int x0, int y0, int radius) {
	int f = 1 - radius;
	int ddF_x = 1;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
	
	vdc_draw_pixel(x0, y0 + radius);
	vdc_draw_pixel(x0, y0 - radius);
	vdc_draw_pixel(x0 + radius, y0);
	vdc_draw_pixel(x0 - radius, y0);
	
	while(x < y) {
		// ddF_x == 2 * x + 1;
		// ddF_y == -2 * y;
		// f == x*x + y*y - radius*radius + 2*x - y + 1;
		if(f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;    
		vdc_draw_pixel(x0 + x, y0 + y);
		vdc_draw_pixel(x0 - x, y0 + y);
		vdc_draw_pixel(x0 + x, y0 - y);
		vdc_draw_pixel(x0 - x, y0 - y);
		vdc_draw_pixel(x0 + y, y0 + x);
		vdc_draw_pixel(x0 - y, y0 + x);
		vdc_draw_pixel(x0 + y, y0 - x);
		vdc_draw_pixel(x0 - y, y0 - x);
	}
}
